unit HrastUnit;

// version: 9
// Copyright  2014 by HrastProgrammer. All rights reserved.
// This unit contains WREN synthesizer modules developed by HrastProgrammer.
// Can be distributed only as part of the complete WREN source package.
//
// version  8: Blue Hell : changed module registration procedure
// version  9: Blue Hell : layout changes
// version 10: Blue Hell : some code changes
// version 11: Blue Hell : added constructors for the Fast/Slow selection

interface

uses

  KnobsUtils, KnobsConversions, Module_Defs, Module_Implementations;

type

  THrastOsc = class( TMod)
  private
    z        : TSignal;
    Phase    : TSignal;
    PrevSync : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                       override;
    procedure   SetDefaults;                                                              override;
    procedure   DoTick;                                                                   override;
    procedure   GatherLights( const Prefix: string; const Callback: TLightsHandler);      override;
    function    InputName ( Index: Integer): string;                                      override;
    function    OutputName( Index: Integer): string;                                      override;
  end;

  THrastSVF = class( TMod)
  private
    Band12 : TSignal;
    Low12  : TSignal;
    Band24 : TSignal;
    Low24  : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                       override;
    procedure   SetDefaults;                                                              override;
    procedure   DoTick;                                                                   override;
    function    InputName ( Index: Integer): string;                                      override;
    function    OutputName( Index: Integer): string;                                      override;
  end;


  procedure CreateTables; // BLue Hell : renamed inintialize to CreateTables and made it public; table re-init is needed
                          //             after changing the system rate



implementation



uses

  Globals, Math, SysUtils;

const

  TableSize = 8192;

var

  Pow2Table : array [ 0 .. TableSize + 1] of TSignal;
  SineTable : array [ 0 .. TableSize + 1] of TSignal;


  function Pow2Freq( Freq: TSignal): TSignal;
  var
    n: Integer;
    f: TSignal;
  begin
    f := Abs( Freq) * TableSize;
    if f > TableSize - 1
    then f := TableSize - 1;
    n := Trunc( f);
    f := f - n;
    Result := Pow2Table[ n];
    Result := Result + f * ( Pow2Table[ n + 1] - Result);
  end;

  function Sine( Freq: TSignal): TSignal;
  var
    n: Integer;
    f: TSignal;
  begin
    f := Abs( Freq) * TableSize;
    if f > TableSize - 1
    then f := TableSize - 1;
    n := Trunc( f);
    f := f - n;
    Result := SineTable[ n];
    Result := Result + f * ( SineTable[ n + 1] - Result);
  end;

{ THrastOsc }


{ ========
  THrastOsc = class( TMod)
  private
    z        : TSignal;
    Phase    : TSignal;
    PrevSync : Boolean;
  public
}

const

  HrastOsc_Freq      =  0;
  HrastOsc_Morph     =  1;
  HrastOsc_Shape     =  2;
  HrastOsc_Bandlimit =  3;
  HrastOsc_PW        =  4;
  HrastOsc_PWM       =  5;
  HrastOsc_PWMInput  =  6;
  HrastOsc_FM        =  7;
  HrastOsc_FMInput   =  8;
  HrastOsc_PM        =  9;
  HrastOsc_PMInput   = 10;
  HrastOsc_Vintage   = 11;
  HrastOsc_SyncInput = 12;
  HrastOsc_Level     = 13;
  HrastOsc_Mute      = 14;
  HrastOsc_Cents     = 15;


    constructor THrastOsc.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   THrastOsc.SetDefaults;
    begin
      FInputs[ HrastOsc_FMInput ] := 0;
      FInputs[ HrastOsc_PMInput ] := 0;
      FInputs[ HrastOsc_PWMInput] := 0;
      FDezipperMap := [
        HrastOsc_Freq,
        HrastOsc_Morph,
        HrastOsc_Shape,
        HrastOsc_Bandlimit,
        HrastOsc_PW,
        HrastOsc_PWM,
        HrastOsc_FM,
        HrastOsc_PM,
        HrastOsc_Level,
        HrastOsc_Mute,
        HrastOsc_Cents
      ];
    end;

    procedure THrastOsc.DoTick;
    var
      DeltaPhase: TSignal;

//      function PolyBLEP( t: TSignal): TSignal;
//      var
//        dt: TSignal;
//      begin
//        dt := FInputs[ HrastOsc_Bandlimit] * DeltaPhase;
//        if t < dt
//        then begin
//          t      := t / dt - 1;
//          Result := - t * t;
//        end
//        else begin
//          if t > 1 - dt
//          then begin
//            t      := ( t - 1) / dt + 1;
//            Result := t * t;
//          end
//          else Result := 0
//        end;
//      end;

    var
      n   : Integer;
      t   : TSignal;
      y   : TSignal;
      ypw : TSignal;
    begin
      if not PrevSync and SignalToLogic( FInputs[HrastOsc_SyncInput])
      then Phase := 0;

      PrevSync := SignalToLogic( FInputs[ HrastOsc_SyncInput]);

      // if SignalToLogic( FInputs[ HrastOsc_Mute])
      // then begin
      //   FOutputs[ 0] := 0;
      //   Exit;
      // end;

      DeltaPhase := Pow2Freq( FInputs[ HrastOsc_Freq] + FInputs[ HrastOsc_Cents] + FInputs[ HrastOsc_FM] * FInputs[ HrastOsc_FMInput]) / System_Rate;

      t := Phase + FInputs[ HrastOsc_PM] * FInputs[ HrastOsc_PMInput];
      n := Trunc( t);
      t := t - n;
      if t < 0
      then t := t + 1;

      Phase := Phase + DeltaPhase;

      n     := Trunc( Phase);
      Phase := Phase - n;
      if Phase < 0
      then Phase := Phase + 1;

      y := 2 * t - 1 - PolyBlep( t, FInputs[ HrastOsc_Bandlimit], DeltaPhase);

      t := t + FInputs[ HrastOsc_PW] + FInputs[ HrastOsc_PWM] * FInputs[ HrastOsc_PWMInput];

      n := Trunc( t);
      t := t - n;
      if t < 0
      then t := t + 1;

      ypw := 2 * t - 1 - PolyBlep( t, FInputs[ HrastOsc_Bandlimit], DeltaPhase);

      t := FInputs[ HrastOsc_Shape];

      if FInputs[ HrastOsc_Vintage] <> 0
      then begin                            // Vintage
        if t > 0.999
        then t := 0.999;
        y := ( y + 1) * 0.5;
        y := y * y;
        y := 2 * y - 1;

        ypw := ( ypw + 1) * 0.5;
        ypw := ypw * ypw;
        ypw := 2 * ypw - 1;
      end;
      z := y - ypw * FInputs[ HrastOsc_Morph] + t * z;

      if t>0.99
      then t:=0.99;

      y := z * ( t - 1);
      FOutputs[ 0] := y * FInputs[ HrastOsc_Level] * SignalToMute( FInputs[ HrastOsc_Mute]);
    end;

    procedure THrastOsc.GatherLights;
    begin
      Callback( Self, Format( '%s/%s/SyncInput/%f', [ Prefix, Name, FInputs[ HrastOsc_SyncInput]], AppLocale));
    end;

    function THrastOsc.InputName;
    begin
      case Index of
        HrastOsc_Freq      : Result := 'Freq';
        HrastOsc_Morph     : Result := 'Morph';
        HrastOsc_Shape     : Result := 'Shape';
        HrastOsc_Bandlimit : Result := 'Bandlimit';
        HrastOsc_PW        : Result := 'PW';
        HrastOsc_PWM       : Result := 'PWM';
        HrastOsc_PWMInput  : Result := 'PWMInput';
        HrastOsc_FM        : Result := 'FM';
        HrastOsc_FMInput   : Result := 'FMInput';
        HrastOsc_PM        : Result := 'PM';
        HrastOsc_PMInput   : Result := 'PMInput';
        HrastOsc_Vintage   : Result := 'Vintage';
        HrastOsc_SyncInput : Result := 'SyncInput';
        HrastOsc_Level     : Result := 'Level';
        HrastOsc_Mute      : Result := 'Mute';
        HrastOsc_Cents     : Result := 'Cents';
        else                 Result := NO_INPUT;
      end
    end;

    function THrastOsc.OutputName;
    begin
      case Index of
        0  : Result := 'Output';
        else Result := NO_OUTPUT;
      end
    end;


{ ========
  THrastSVF = class( TMod)
  private
    Band12 : TSignal;
    Low12  : TSignal;
    Band24 : TSignal;
    Low24  : TSignal;
  public
}

const

  HrastSVF_Input       =  0;
  HrastSVF_Cutoff      =  1;
  HrastSVF_Reso        =  2;
  HrastSVF_CutoffInput =  3;
  HrastSVF_CutoffMod   =  4;
  HrastSVF_ResoInput   =  5;
  HrastSVF_ResoMod     =  6;
  HrastSVF_Gain        =  7;
  HrastSVF_Drive       =  8;
  HrastSVF_LowPass     =  9;
  HrastSVF_BandPass    = 10;
  HrastSVF_HighPass    = 11;
  HrastSVF_Level12     = 12;
  HrastSVF_Level24     = 13;

  HrastSVF_OutputLP    =  0;
  HrastSVF_OutputBP    =  1;
  HrastSVF_OutputHP    =  2;
  HrastSVF_OutputBR    =  3;
  HrastSVF_Output12    =  4;
  HrastSVF_Output24    =  5;


    constructor THrastSVF.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   THrastSVF.SetDefaults;
    begin
      FInputs[ HrastSVF_Input      ] := 0;
      FInputs[ HrastSVF_CutoffInput] := 0;
      FInputs[ HrastSVF_ResoInput  ] := 0;
      FDezipperMap:=[
        HrastSVF_Cutoff,
        HrastSVF_Reso,
        HrastSVF_CutoffMod,
        HrastSVF_ResoMod,
        HrastSVF_Gain,
        HrastSVF_Drive,
        HrastSVF_LowPass,
        HrastSVF_BandPass,
        HrastSVF_HighPass,
        HrastSVF_Level12,
        HrastSVF_Level24
      ]
    end;

    procedure THrastSVF.DoTick;
    var
      Oversample: Integer;
      f,fb,y,High12,High24,Output12,Output24: TSignal;
    begin
      f  := Pow2Freq( FInputs[ HrastSVF_Cutoff] + FInputs[ HrastSVF_CutoffMod] * FInputs[ HrastSVF_CutoffInput]) / System_Rate_2;
      f  := 2 * Sine( f);
      fb := Clip( 1 - ( FInputs[ HrastSVF_Reso] + FInputs[ HrastSVF_ResoMod] * FInputs[ HrastSVF_ResoInput]), 0, 1);

      // if fb < 0
      // then fb := 0
      // else if fb > 1
      // then fb := 1;

      y := FInputs[ HrastSVF_Input] * FInputs[ HrastSVF_Gain];

      // if y = 0
      // then y := 1e-10; // BlueHell added this, had division by zero in some patch

      y := y / ( Abs( y) + ( 1 - Clip( FInputs[ HrastSVF_Drive], 0, 0.999)));

      for Oversample := 1 to 2
      do begin
        High12   := y - fb * Band12 - Low12;
        Band12   := f * High12 + Band12;
        Low12    := f * Band12 + Low12;
        Low12    := Low12 * 1.0001 / ( 0.0001 * Abs( Low12) + 1);
        Output12 := Low12 * FInputs[ HrastSVF_LowPass] + Band12 * FInputs[ HrastSVF_BandPass] + High12 * FInputs[ HrastSVF_HighPass];

        High24   := Output12 - fb * Band24 - Low24;
        Band24   := f * High24 + Band24;
        Low24    := f * Band24 + Low24;
        Low24    := Low24 * 1.0001 / ( 0.01 * Abs( Low24) + 1);
      end;
      Output24 := Low24 * FInputs[ HrastSVF_LowPass] + Band24 * FInputs[ HrastSVF_BandPass] + High24 * FInputs[ HrastSVF_HighPass];
      FOutputs[ HrastSVF_OutputLP] := Low12;
      FOutputs[ HrastSVF_OutputBP] := Band12;
      FOutputs[ HrastSVF_OutputHP] := High12;
      FOutputs[ HrastSVF_OutputBR] := Low12 + High12;
      FOutputs[ HrastSVF_Output12] := Output12 * FInputs[ HrastSVF_Level12];
      FOutputs[ HrastSVF_Output24] := Output24 * FInputs[ HrastSVF_Level24];
    end;

    function THrastSVF.InputName;
    begin
      case Index of
        HrastSVF_Input       : Result := 'Input';
        HrastSVF_Cutoff      : Result := 'Cutoff';
        HrastSVF_Reso        : Result := 'Reso';
        HrastSVF_CutoffInput : Result := 'CutoffInput';
        HrastSVF_CutoffMod   : Result := 'CutoffMod';
        HrastSVF_ResoInput   : Result := 'ResoInput';
        HrastSVF_ResoMod     : Result := 'ResoMod';
        HrastSVF_Gain        : Result := 'Gain';
        HrastSVF_Drive       : Result := 'Drive';
        HrastSVF_LowPass     : Result := 'LowPass';
        HrastSVF_BandPass    : Result := 'BandPass';
        HrastSVF_HighPass    : Result := 'HighPass';
        HrastSVF_Level12     : Result := 'Level12';
        HrastSVF_Level24     : Result := 'Level24';
        else                   Result := NO_INPUT;
      end
    end;

    function THrastSVF.OutputName;
    begin
      case Index of
        HrastSVF_OutputLP : Result := 'OutputLP';
        HrastSVF_OutputBP : Result := 'OutputBP';
        HrastSVF_OutputHP : Result := 'OutputHP';
        HrastSVF_OutputBR : Result := 'OutputBR';
        HrastSVF_Output12 : Result := 'Output12';
        HrastSVF_Output24 : Result := 'Output24';
        else                Result := NO_OUTPUT;
      end
    end;



  procedure CreateTables;
  var
    i: Integer;
    x: TSignal;
  begin
    for i := 0 to TableSize
    do Pow2Table[ i] := UnitsToFrequency( i / TableSize);
    Pow2Table[ TableSize + 1] := Pow2Table[ TableSize];

    x := Pi / TableSize;
    for i := 0 to TableSize - 1
    do SineTable[ i] := Sin( x * i);
    SineTable[ TableSize div 2] := 0;
    SineTable[ TableSize      ] := SineTable[ 0];
    SineTable[ TableSize + 1  ] := SineTable[ 1];
  end;

initialization

  CreateTables;

end.
